@@ -23,10 +23,12 @@
 #pragma warning(disable: 4996)
 #endif
 
+#define INITIAL_DITHER_SEED 0x12345000
+
 static int32_t smpShiftValue;
-static uint32_t oldAudioFreq, tickTimeLenInt;
+static uint32_t oldAudioFreq, tickTimeLenInt, randSeed = INITIAL_DITHER_SEED;
 static uint64_t tickTimeLenFrac;
-static float fAudioNormalizeMul, fSqrtPanningTable[256+1];
+static float fSqrtPanningTable[256+1], fAudioNormalizeMul, fPrngStateL, fPrngStateR;
 static voice_t voice[MAX_CHANNELS * 2];
 
 // globalized
@@ -100,11 +102,11 @@ void setAudioAmp(int16_t amp, int16_t masterVol, bool bitDepth32Flag)
 	amp = CLAMP(amp, 1, 32);
 	masterVol = CLAMP(masterVol, 0, 256);
 
-	double dAmp = (amp * masterVol) / (32.0 * 256.0);
+	float fAmp = (amp * masterVol) / (32.0f * 256.0f);
 	if (!bitDepth32Flag)
-		dAmp *= 32768.0;
+		fAmp *= 32768.0f;
 
-	fAudioNormalizeMul = (float)dAmp;
+	fAudioNormalizeMul = fAmp;
 }
 
 void decreaseMasterVol(void)
@@ -191,17 +193,15 @@ void audioSetInterpolationType(uint8_t interpolationType)
 	// set sinc LUT pointers
 	if (config.interpolation == INTERPOLATION_SINC8)
 	{
-		fSinc_1 = fSinc8_1;
-		fSinc_2 = fSinc8_2;
-		fSinc_3 = fSinc8_3;
+		for (int32_t i = 0; i < SINC_KERNELS; i++)
+			fSinc[i] = fSinc8[i];
 
 		audio.sincInterpolation = true;
 	}
 	else if (config.interpolation == INTERPOLATION_SINC16)
 	{
-		fSinc_1 = fSinc16_1;
-		fSinc_2 = fSinc16_2;
-		fSinc_3 = fSinc16_3;
+		for (int32_t i = 0; i < SINC_KERNELS; i++)
+			fSinc[i] = fSinc16[i];
 
 		audio.sincInterpolation = true;
 	}
@@ -234,11 +234,10 @@ static void voiceUpdateVolumes(int32_t i, uint8_t status)
 
 	// now we need to handle volume ramping
 
-	const bool voiceSampleTrigger = !!(status & IS_Trigger);
-
-	if (voiceSampleTrigger)
+	const bool voiceTriggerFlag = !!(status & CS_TRIGGER_VOICE);
+	if (voiceTriggerFlag)
 	{
-		// sample is about to start, ramp out/in at the same time
+		// voice is about to start, ramp out/in at the same time
 
 		if (v->fCurrVolumeL > 0.0f || v->fCurrVolumeR > 0.0f)
 		{
@@ -262,7 +261,7 @@ static void voiceUpdateVolumes(int32_t i, uint8_t status)
 		v->fCurrVolumeL = v->fCurrVolumeR = 0.0f;
 	}
 
-	if (!voiceSampleTrigger && v->fTargetVolumeL == v->fCurrVolumeL && v->fTargetVolumeR == v->fCurrVolumeR)
+	if (!voiceTriggerFlag && v->fTargetVolumeL == v->fCurrVolumeL && v->fTargetVolumeR == v->fCurrVolumeR)
 	{
 		v->volumeRampLength = 0; // no ramp needed for now
 	}
@@ -272,7 +271,7 @@ static void voiceUpdateVolumes(int32_t i, uint8_t status)
 		const float fVolumeRDiff = v->fTargetVolumeR - v->fCurrVolumeR;
 
 		float fRampLengthMul;
-		if (status & IS_QuickVol) // duration of 5ms
+		if (status & CS_USE_QUICK_VOLRAMP) // 5ms duration
 		{
 			v->volumeRampLength = audio.quickVolRampSamples;
 			fRampLengthMul = audio.fQuickVolRampSamplesMul;
@@ -365,54 +364,78 @@ void updateVoices(void)
 
 		ch->status = 0;
 
-		if (status & IS_Vol)
+		if (status & CS_UPDATE_VOL)
 		{
 			v->fVolume = ch->fFinalVol; // 0.0f .. 1.0f
-			v->scopeVolume = (uint8_t)(v->fVolume * 255.0f);
+			v->scopeVolume = (uint8_t)((ch->fFinalVol * (SCOPE_HEIGHT*4.0f)) + 0.5f);
 		}
 
-		if (status & IS_Pan)
+		if (status & CS_UPDATE_PAN)
 			v->panning = ch->finalPan;
 
-		if (status & (IS_Vol + IS_Pan))
+		if (status & (CS_UPDATE_VOL + CS_UPDATE_PAN)) // for vol and/or pan updates
 			voiceUpdateVolumes(i, status);
 
-		if (status & IS_Period)
+		if (status & CF_UPDATE_PERIOD)
 		{
-			const double dVoiceHz = dPeriod2Hz(ch->finalPeriod);
+			v->delta = period2VoiceDelta(ch->finalPeriod);
 
-			// set voice delta
-			v->delta = (int64_t)((dVoiceHz * audio.dHz2MixDeltaMul) + 0.5); // Hz -> fixed-point delta (rounded)
 			if (audio.sincInterpolation)
 			{
-				// decide which sinc LUT to use according to the resampling ratio
 				if (v->delta <= sincRatio1)
-					v->fSincLUT = fSinc_1;
+					v->fSincLUT = fSinc[0];
 				else if (v->delta <= sincRatio2)
-					v->fSincLUT = fSinc_2;
+					v->fSincLUT = fSinc[1];
 				else
-					v->fSincLUT = fSinc_3;
+					v->fSincLUT = fSinc[2];
 			}
 		}
 
-		if (status & IS_Trigger)
+		if (status & CS_TRIGGER_VOICE)
 			voiceTrigger(i, ch->smpPtr, ch->smpStartPos);
 	}
 }
 
+void resetAudioDither(void)
+{
+	randSeed = INITIAL_DITHER_SEED;
+	fPrngStateL = fPrngStateR = 0.0f;
+}
+
+static inline int32_t random32(void)
+{
+	// LCG 32-bit random
+	randSeed *= 134775813;
+	randSeed++;
+
+	return (int32_t)randSeed;
+}
+
 static void sendSamples16BitStereo(void *stream, uint32_t sampleBlockLength)
 {
+	int32_t out32;
+	float fOut, fPrng;
+
 	int16_t *streamPtr16 = (int16_t *)stream;
 	for (uint32_t i = 0; i < sampleBlockLength; i++)
 	{
-		int32_t L = (int32_t)(audio.fMixBufferL[i] * fAudioNormalizeMul);
-		int32_t R = (int32_t)(audio.fMixBufferR[i] * fAudioNormalizeMul);
+		// left channel - 1-bit triangular dithering
+		fPrng = (float)random32() * (1.0f / (UINT32_MAX+1.0f)); // -0.5f .. 0.5f
+		fOut = audio.fMixBufferL[i] * fAudioNormalizeMul;
+		fOut = (fOut + fPrng) - fPrngStateL;
+		fPrngStateL = fPrng;
+		out32 = (int32_t)fOut;
+		CLAMP16(out32);
+		*streamPtr16++ = (int16_t)out32;
 
-		CLAMP16(L);
-		CLAMP16(R);
-
-		*streamPtr16++ = (int16_t)L;
-		*streamPtr16++ = (int16_t)R;
+		// right channel - 1-bit triangular dithering
+		fPrng = (float)random32() * (1.0f / (UINT32_MAX+1.0f)); // -0.5f .. 0.5f
+		fOut = audio.fMixBufferR[i] * fAudioNormalizeMul;
+		fOut = (fOut + fPrng) - fPrngStateR;
+		fPrngStateR = fPrng;
+		out32 = (int32_t)fOut;
+		CLAMP16(out32);
+		*streamPtr16++ = (int16_t)out32;
 
 		// clear what we read from the mixing buffer
 		audio.fMixBufferL[i] = audio.fMixBufferR[i] = 0.0f;
@@ -421,14 +444,20 @@ static void sendSamples16BitStereo(void *stream, uint32_t sampleBlockLength)
 
 static void sendSamples32BitFloatStereo(void *stream, uint32_t sampleBlockLength)
 {
+	float fOut;
+
 	float *fStreamPtr32 = (float *)stream;
 	for (uint32_t i = 0; i < sampleBlockLength; i++)
 	{
-		const float fL = audio.fMixBufferL[i] * fAudioNormalizeMul;
-		const float fR = audio.fMixBufferR[i] * fAudioNormalizeMul;
+		// left channel
+		fOut = audio.fMixBufferL[i] * fAudioNormalizeMul;
+		fOut = CLAMP(fOut, -1.0f, 1.0f);
+		*fStreamPtr32++ = fOut;
 
-		*fStreamPtr32++ = CLAMP(fL, -1.0f, 1.0f);
-		*fStreamPtr32++ = CLAMP(fR, -1.0f, 1.0f);
+		// right channel
+		fOut = audio.fMixBufferR[i] * fAudioNormalizeMul;
+		fOut = CLAMP(fOut, -1.0f, 1.0f);
+		*fStreamPtr32++ = fOut;
 
 		// clear what we read from the mixing buffer
 		audio.fMixBufferL[i] = audio.fMixBufferR[i] = 0.0f;
@@ -770,7 +799,7 @@ static void fillVisualsSyncBuffer(void)
 		c->smpStartPos = s->smpStartPos;
 
 		c->pianoNoteNum = 255; // no piano key
-		if (songPlaying && ui.instEditorShown && (c->status & IS_Period) && !s->keyOff)
+		if (songPlaying && ui.instEditorShown && (c->status & CF_UPDATE_PERIOD) && !s->keyOff)
 		{
 			const int32_t note = getPianoKey(s->finalPeriod, s->finetune, s->relativeNote);
 			if (note >= 0 && note <= 95)
@@ -794,12 +823,17 @@ static void fillVisualsSyncBuffer(void)
 static void SDLCALL audioCallback(void *userdata, Uint8 *stream, int len)
 {
 	if (editor.wavIsRendering)
+	{
+		memset(stream, 0, len);
 		return;
+	}
 
 	len >>= smpShiftValue; // bytes -> samples
 	if (len <= 0)
 		return;
 
+	audio.callbackOngoing = true;
+
 	int32_t bufferPosition = 0;
 
 	uint32_t samplesLeft = len;
@@ -845,6 +879,8 @@ static void SDLCALL audioCallback(void *userdata, Uint8 *stream, int len)
 	else
 		sendSamples32BitFloatStereo(stream, len);
 
+	audio.callbackOngoing = false;
+
 	(void)userdata;
 }
 
@@ -1064,4 +1100,6 @@ void closeAudio(void)
 	}
 
 	freeAudioBuffers();
+
+	audio.callbackOngoing = false;
 }
